#define _BSD_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <linux/limits.h>
#include <dlfcn.h>
#include <sqlite3.h>

#define sqlite3_open (*P_sqlite3_open)
static int (*P_sqlite3_open)(const char *, sqlite3 **);
#define sqlite3_exec (*P_sqlite3_exec)
static int (*P_sqlite3_exec)(sqlite3 *, const char *, int (*)(void *, int, char **, char **), void *, char **);
#define sqlite3_close (*P_sqlite3_close)
static int (*P_sqlite3_close)(sqlite3 *);

int main(void);
static void geturls(const char *appname, const char *profiledir);
static int callback(void *arg, int num, char **val, char **key);

struct entry {
   char *browser;
   char *url;
   char *title;
   time_t time;
   struct entry *next;
};

static struct entry *list = NULL, *listp = NULL;

int main(void)
{
   char profiles[PATH_MAX], profilepath[PATH_MAX], line[1024], *profiledir = NULL;
   FILE *fp = NULL;
   int relative = 1;

   do {
      if(snprintf(profiles, sizeof(profiles), "%s/.mozilla/firefox/profiles.ini", getenv("HOME")) >= sizeof(profiles)) break;
      if(!(fp = fopen(profiles, "r"))) break;
      while(!feof(fp)) {
         if(!fgets(line, sizeof(line), fp)) break;
         if(!strncmp(line, "IsRelative=", strlen("IsRelative="))) relative = atoi(line + strlen("IsRelative="));
         if(!strncmp(line, "Path=", strlen("Path="))) {
            profiledir = line + strlen("Path=");
            while(profiledir[strlen(profiledir) - 1] == '\n') profiledir[strlen(profiledir) - 1] = '\0';
            if(relative) {
               if(snprintf(profilepath, sizeof(profilepath), "%s/.mozilla/firefox/%s", getenv("HOME"), profiledir) >= sizeof(profilepath)) continue;
               profiledir = profilepath;
            }
            geturls("firefox", profiledir);
         }
      }
   } while(0);
   if(fp) fclose(fp);
   fp = NULL;

   for(listp = list; listp;) {
      printf("%u %s %s (%s)\n", listp->time, listp->browser, listp->url, listp->title);

      list = listp;
      listp = listp->next;
      if(list->browser) free(list->browser);
      if(list->url) free(list->url);
      if(list->title) free(list->title);
      free(list);
   }
   return 0;
}

static void geturls(const char *appname, const char *profiledir)
{
   void *h1 = NULL;
   sqlite3 *db = NULL;
   FILE *fp = NULL;
   char sql[] = "SELECT url, title, visit_date/1000000 FROM moz_historyvisits JOIN moz_places ON moz_historyvisits.place_id = moz_places.id WHERE visit_date > %u ORDER BY moz_historyvisits.id";
   char compatibility[PATH_MAX], places[PATH_MAX], libpath[PATH_MAX], line[1024], *libdir = NULL, query[sizeof(sql) + 32];;

   do {
      do {
         if(snprintf(compatibility, sizeof(compatibility), "%s/compatibility.ini", profiledir) >= sizeof(compatibility)) break;
         if(!(fp = fopen(compatibility, "r"))) break;
         while(!feof(fp)) {
            if(!fgets(line, sizeof(line), fp)) break;
            if(!strncmp(line, "LastAppDir=", strlen("LastAppDir="))) {
               libdir = line + strlen("LastAppDir=");
               while(libdir[strlen(libdir) - 1] == '\n') libdir[strlen(libdir) - 1] = '\0';
               break;
            }
         }
      } while(0);
      if(fp) fclose(fp);

      do {
         if(libdir) {
            if(snprintf(libpath, sizeof(libpath), "%s/libmozsqlite3.so", libdir) >= sizeof(libpath)) break;
            if((h1 = dlopen(libpath, RTLD_NOW))) break;
         }
         if(snprintf(libpath, sizeof(libpath), "/usr/lib/%s/libmozsqlite3.so", appname) >= sizeof(libpath)) break;
         if((h1 = dlopen(libpath, RTLD_NOW))) break;
      } while(0);
      if(!h1) break;

      if(!(*(void **)(&P_sqlite3_open) = dlsym(h1, "sqlite3_open"))) break;
      if(!(*(void **)(&P_sqlite3_exec) = dlsym(h1, "sqlite3_exec"))) break;
      if(!(*(void **)(&P_sqlite3_close) = dlsym(h1, "sqlite3_close"))) break;

      if(snprintf(places, sizeof(places), "%s/places.sqlite", profiledir) >= sizeof(places)) break;
      if(snprintf(query, sizeof(query), sql, 0) >= sizeof(query)) break;

      if(sqlite3_open(places, &db) != SQLITE_OK) break;
      if(sqlite3_exec(db, query, callback, (void *)appname, NULL) != SQLITE_OK) break;
   } while(0);
   if(db) sqlite3_close(db);
   if(h1) dlclose(h1);

   return;
}

static int callback(void *arg, int num, char **val, char **key)
{
   do {
      if(num != 3) break;

      if(!list) {
         if(!(list = malloc(sizeof(struct entry)))) break;
         listp = list;
      } else {
         if(!(listp->next = malloc(sizeof(struct entry)))) break;
         listp = listp->next;
      }

      listp->next = NULL;
      listp->browser = strdup(arg ? arg : "");
      listp->url = strdup(val[0] ? val[0] : "");
      listp->title = strdup(val[1] ? val[1] : "");
      listp->time = (time_t)(val[2] ? atoll(val[2]) : 0);
   } while(0);

   return 0;
}
